package Source::Repository::Mass::SkipPkg::ImportSkipMassRebuild;

use strict;
use warnings;
use Carp;
use File::Basename;
use RPM::Header;
use RPM::Vercmp;

# due to Source::Repository::Mass::version_comparator
use Source::Repository::Mass;
use Source::Package::Comparators::Raw;
use Source::Repository::Mass::SkipPkg;
our @ISA=qw/Source::Repository::Mass::SkipPkg/;
push @Source::Repository::Mass::PkgFilter::PKG_SKIP_CLASSES, __PACKAGE__;

our $verbose=1;
our $arg_enable_filter=0;
our $arg_skip_rebuilds=1;
our $arg_rebuilds_depth=0; # number of rebuilds to skip
our $arg_skip_shallow=1;

our @LONGOPT = (
    "changelog-rebuild-on"	=> sub {$arg_enable_filter=1},
    "changelog-rebuild-off"	=> sub {$arg_enable_filter=0},
    "changelog-rebuild-select"	=> sub {$arg_enable_filter=1,$arg_skip_rebuilds=0},
    "changelog-rebuild-skip"	=> sub {$arg_enable_filter=1,$arg_skip_rebuilds=1},
    "changelog-rebuild-depth=s"	=> \$arg_rebuilds_depth,
    "changelog-rebuild-include-shallow!"	=> \$arg_skip_shallow,
    );

sub print_longopt {
    print "  `Ignore changelog Mass Rebuild' options:
	--changelog-rebuild-skip  enable skippkg of mass rebuilds
	--changelog-rebuild-depth <num> number of mass rebuilds to skip
	--changelog-rebuild-include-shallow skip shallow changelogs too (default)
	--no-changelog-rebuild-include-shallow don't skip shallow changelogs too
";
}

sub pkgfilter {
    my ($self,$pkgpair)=@_;
    return 0 unless $arg_enable_filter;
    my $filter_call=$self->pkgfilter_if_skip_rebuilds($pkgpair);
    if ($arg_skip_rebuilds) {
	print STDERR "note: skipped Mass_Rebuild ",$pkgpair->originname,"\n" if $filter_call and $verbose;
	return $filter_call;
    } else {
	print STDERR "note: selected Mass_Rebuild ",$pkgpair->originname,"\n" if not $filter_call and $verbose;
	return ! $filter_call;
    }
}


sub pkgfilter_if_skip_rebuilds {
    my ($self,$pkgpair)=@_;
    my $origin=$pkgpair->origin;
    my $dest=$pkgpair->dest;
    return 0 unless $dest;
    unless ($origin) {
	my $originname=$pkgpair->originname;
	my $destname=$pkgpair->destname;
	print STDERR "Oops! no origin found for $originname ($destname). Skipping\n";
	return 1;
    }
    my $origin_version=$origin->version;
    my $dest_version=$dest->version;
    my $origin_release=$origin->release;
    my $dest_release=$dest->release;
    my $lazyfunc_get_changelog=__::LazyFuncCall->new(\&__get_changelog_lists_from_rpmfile,$origin->file);
    my $lazyfunc_is_same_version=__::LazyFuncCall->new($Source::Repository::Mass::version_comparator,$origin,$dest);
    return &is_changelog_has_nothing_but_rebuilds($origin_version,$dest_version,$origin_release,$dest_release,$lazyfunc_get_changelog,$lazyfunc_is_same_version);
}

#------------------------------------------
#    public funcs (shared)
#------------------------------------------

# the code below does not use origin/dest and can be shared as CLI with R::S::Transformation
sub is_changelog_has_nothing_but_rebuilds {
    my ($origin_version,$dest_version,$origin_release,$dest_release,$lazyfunc_get_changelog,$lazyfunc_is_same_version)=@_;
    return 0 if $lazyfunc_is_same_version->call()!=0
	# hack; we should ignore epoch if version is the same; epoch can be dropped during import,
	# an example is aspell-pt_BR epoch 50 dropped in aspell.pl
	# TODO: better use version part of version comparator
	and $origin_version ne $dest_version;
    my $altoriginrel=&__get_origin_release_from_alt($dest_release);
    return 0 if ! defined($altoriginrel);
    my ($depth,$changelog_depth)=&__calculate_rebuild_depth($origin_version, $altoriginrel,$lazyfunc_get_changelog);
    my $originrel=&Source::Package::Comparators::Raw::import_trim_release($origin_release);
    print STDERR 'DEBUG: skip MassRebuild: altoriginrel=',$altoriginrel,' originrel=',$originrel,' depth=',$depth,"\n" if $verbose>2;
    return 0 if $depth==0;
    return 0 if $arg_rebuilds_depth and $depth > $arg_rebuilds_depth;
    return 1 if $arg_skip_shallow and $depth == $changelog_depth;
    return 0 if &__number($originrel)-&__number($altoriginrel) > $depth;
    return 1;
}

#------------------------------------------
#    private funcs (Mass)
#------------------------------------------
sub __get_changelog_lists_from_rpmfile {
    my ($originfile)=@_;
    die "Oops! file $originfile not found" if ! -e $originfile;
    my $rh=RPM::Header->new($originfile);
    die "Oops! file $originfile: RPM::Header failed" if ! $rh;
    my $changelognameref=$rh->{CHANGELOGNAME};
    my $changelogtextref=$rh->{CHANGELOGTEXT};
    return ($changelognameref,$changelogtextref);
}

#------------------------------------------
#    private funcs (shared)
#------------------------------------------

sub __calculate_rebuild_depth {
    my ($origin_version,$altoriginrel,$lazyfunc_get_changelog)=@_;
    my ($changelognameref,$changelogtextref)=$lazyfunc_get_changelog->call();

    unless ($changelognameref or $changelogtextref) {
	print STDERR "WARNING: ImportSkipMassRebuild: empty changelog", "\n";
	# can't get changelog at all :(
	return 0,0;
    }
    my $depth=0;
    my ($old_ver,$old_rel);
    my $changelog_depth=$#{$changelognameref}+1;
    for (my $i=0; $i<@$changelognameref; $i++) {
	my $changelogheader=$changelognameref->[$i];
	my $changelogtext=$changelogtextref->[$i];
	my ($ver,$rel)=&__get_epochversion_release($changelogheader);
	unless (defined $rel) {
	    if (defined $old_rel) {
		$rel=$old_rel;
	    } else {
		$rel=0;
	    }
	}
	# ignore epoch
	$ver=~s/^\d+://;
	warn 'DEBUG: log ver=',$ver,' altver=',$origin_version,' rel=',$rel,' alt=',$altoriginrel,' power=',scalar @$changelognameref,"\n" if $verbose>4;
	($old_ver,$old_rel)=($ver, $rel);
	return $depth,$changelog_depth if $ver ne $origin_version;
	return $depth,$changelog_depth if RPM::Vercmp::evrcmp($rel,$altoriginrel)<=0;
	my $is_rebuild=&__is_rebuild($changelogheader,$changelogtext);
	if (!$is_rebuild) {
	    warn 'DEBUG: non-rebuild log: depth=', $depth, ' LOG=',$changelogtext,"\n" if $verbose>4;
	    return $depth,$changelog_depth;
	} else {
	    warn 'DEBUG: rebuild log: depth=', $depth, ' LOG=',$changelogtext,"\n" if $verbose>4;
	}
	$depth++;
    }
    return $depth, $changelog_depth;
}

sub __get_epochversion_release {
    my ($changelogname)=@_;
    # 'Fedora Release Engineering <releng@fedoraproject.org> - 0.95-13'
    if ($changelogname=~/-\s+([^-]+)-([^-]+)$/) {
	return ($1,$2);
    # bad changelog, found in rpmsphere
    # Wei-Lun Chao <bluebat@member.fsf.org> - 5.0.2408
    } elsif ($changelogname=~/-\s+([^-]+)$/) {
	warn 'WARNING: ImportSkipMassRebuild: bad release in changelogname: ',$changelogname;
	return ($1,undef);
    # umeabot <umeabot> 1.0.4-7.mga7
    } elsif ($changelogname=~/\>\s+([^-]+)-([^-]+)$/) {
	return ($1,$2);
    } else {
	warn 'WARNING: ImportSkipMassRebuild: bad changelogname: ',$changelogname;
	return (-1,0);
    }
}

sub __is_rebuild {
    my ($changelogname, $changelogtext)=@_;
    # note: umeabot is 2-liner
    # umeabot <umeabot> 1.0.4-7.mga7 # Mageia 7 Mass Rebuild
    return 1 if $changelogname=~/umeabot \<umeabot\>/ and $changelogtext=~m!Mageia [\d\.]+ Mass Rebuild!;
    # remove mageia autorevision if any
    #+ Revision: 1747459
    $changelogtext=~s/^\+\s+Revision:\s+\d+\s*\n//;
    # 'Fedora Release Engineering <releng@fedoraproject.org> - '
    # 'RPM Fusion Release Engineering <leigh123linux@gmail.com> - '
    return 1 if $changelogname=~/ Release Engineering/ and $changelogtext=~m!Rebuilt for! and $changelogtext=~m!https://fedoraproject.org/wiki/!;
    # 2+-liner: not a rebuild line in fedora
    return 0 if $changelogname=~/.\n\s*\S/s;
    #- Rebuild for readline 8.0
    #- Rebuild fo new imlib
    #- Rebuild for SuperLU-6.0
    return 1 if $changelogtext=~m!^- [Rr]ebuil[dt] for? \S+ \S+!;
    return 1 if $changelogtext=~m!^- [Rr]ebuil[dt] for \S+\s*$!;
    return 1 if $changelogtext=~m!^- Rebuild \(\S+\)\s*$!;
    return 1 if $changelogtext=~m!^- Perl [\d\.]+ rebuild\s*$!;
    return 0;
}

sub __get_origin_release_from_alt {
    my $altrel=shift;
    my $originrel;
    if ($altrel=~/^(?:[^_]+)_(.*?)(?:jpp\d+)?$/) {
	$originrel=$1;
    }
    return $originrel;
}

sub __number {
    my $rel=shift;
    $rel=$1 if $rel=~m/^(\d+(?:\.\d+)?)/;
    return $rel;
}

package __::LazyFuncCall;

sub new {
    my ($class,$func,@args)=@_;
    return bless {-func=>$func, -args=>\@args}, $class;
}

sub call {
    my $self=shift;
    return &{$self->{-func}}(@{$self->{-args}},@_);
}

1;

__END__
'- Rebuilt fo new imlib',
'- Rebuilt for https://fedoraproject.org/wiki/Fedora_28_Mass_Rebuild',
'- Rebuilt for https://fedoraproject.org/wiki/Fedora_27_Binutils_Mass_Rebuild',
'- Rebuilt for https://fedoraproject.org/wiki/Fedora_27_Mass_Rebuild',
'- Rebuilt for https://fedoraproject.org/wiki/Fedora_26_Mass_Rebuild',
'- Rebuilt for https://fedoraproject.org/wiki/Fedora_24_Mass_Rebuild',
'- Rebuilt for https://fedoraproject.org/wiki/Fedora_23_Mass_Rebuild',
'- Rebuilt for https://fedoraproject.org/wiki/Fedora_21_22_Mass_Rebuild',
'- Rebuilt for https://fedoraproject.org/wiki/Fedora_21_Mass_Rebuild',
'- Rebuilt for https://fedoraproject.org/wiki/Fedora_19_Mass_Rebuild',
'- Rebuilt for https://fedoraproject.org/wiki/Fedora_18_Mass_Rebuild',
'- Rebuilt for https://fedoraproject.org/wiki/Fedora_17_Mass_Rebuild',
'- Second attempt - Rebuilt for
  https://fedoraproject.org/wiki/Fedora_33_Mass_Rebuild'
* Sun Sep 23 2018 umeabot <umeabot> 1.0.4-7.mga7
  (not released yet)
+ Revision: 1297284
- Mageia 7 Mass Rebuild
- Rebuild for readline 8.0
- Rebuilt for Python 3.8
- Perl 5.32 rebuild
- Rebuild (geos)
